package com.mindoo.ftp.connection.nsf;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.List;
import java.util.Vector;
import java.util.logging.Level;

import lotus.domino.Database;
import lotus.domino.Document;
import lotus.domino.Name;
import lotus.domino.NotesException;
import lotus.domino.NotesFactory;
import lotus.domino.NotesThread;
import lotus.domino.Session;
import lotus.domino.View;

import org.apache.ftpserver.impl.PassivePorts;
import org.apache.mina.filter.firewall.Subnet;

import com.mindoo.ftp.config.nsf.FTPDbLocator;
import com.mindoo.ftp.connection.DataConnectionConfigurationWrapper;
import com.mindoo.ftp.internal.FTPActivator;

/**
 * Connection configuration that reads part of the config data from an
 * NSF database.
 * 
 * @author Karsten Lehmann
 */
public class NSFDataConnectionConfiguration extends DataConnectionConfigurationWrapper {
	private PassivePorts m_passivePorts;
	private String m_passiveAddress;
	private String m_passiveExternalAddress;
	private Integer m_idleTime;
	private Integer m_ftpPort;
	private boolean m_enabled;
	private List<Subnet> m_blockedSubnets;
	private List<InetAddress> m_blockedAddresses;
	
	public NSFDataConnectionConfiguration() {
		super();
		initConfiguration();
	}

	@Override
	public String getPassivePorts() {
		if (m_passivePorts!=null)
			return m_passivePorts.toString();

		return super.getPassivePorts();
	}

	@Override
	public synchronized void releasePassivePort(int port) {
		if (m_passivePorts!=null) {
			m_passivePorts.releasePort(port);

			notify();
		}
		else {
			super.releasePassivePort(port);
		}
	}

	@Override
	public synchronized int requestPassivePort() {
		if (m_passivePorts!=null) {
			int dataPort = -1;
			int loopTimes = 2;
			Thread currThread = Thread.currentThread();

			while ((dataPort == -1) && (--loopTimes >= 0)
					&& (!currThread.isInterrupted())) {

				// search for a free port
				dataPort = m_passivePorts.reserveNextPort();

				// no available free port - wait for the release notification
				if (dataPort == -1) {
					try {
						FTPActivator.log(Level.INFO, "We're waiting for a passive port, might be stuck", null);
						wait();
					} catch (InterruptedException ex) {
					}
				}
			}
			return dataPort;
		}
		else {
			return super.requestPassivePort();
		}
	}

	@Override
	public String getPassiveAddress() {
		if (m_passiveAddress!=null)
			return m_passiveAddress;

		return super.getPassiveAddress();
	}

	@Override
	public String getPassiveExernalAddress() {
		if (m_passiveExternalAddress!=null)
			return m_passiveExternalAddress;

		return super.getPassiveExernalAddress();
	}

	@Override
	public int getIdleTime() {
		if (m_idleTime!=null)
			return m_idleTime;

		return super.getIdleTime();
	}

	public int getPort() {
		return m_ftpPort==null ? 2221 : m_ftpPort.intValue();
	}

	public boolean isEnabled() {
		return m_enabled;
	}
	
	/**
	 * Load configuration for this server from the config db
	 */
	private void initConfiguration() {
		NotesThread.sinitThread();
		Session session = null;
		try {
			session=NotesFactory.createSession();
			Database configDb=FTPDbLocator.getConfigDatabase(session);
			if (configDb==null) {
				FTPActivator.log(Level.WARNING, "No FTP server configuration database found!", null);
				return;
			}
			View lkView=configDb.getView("($lkConfigByServer)");
			if (lkView==null) {
				FTPActivator.log(Level.WARNING, "View ($lkConfigByServer) not found in FTP server configuration database", null);
				return;
			}

			String serverNameAsString=session.getServerName();
			Name serverName=session.createName(serverNameAsString);
			String serverNameAbbr=serverName.getAbbreviated();
			serverName.recycle();
			Document configDoc=lkView.getDocumentByKey(serverNameAbbr, true);
			if (configDoc==null) {
				m_enabled=false;
				FTPActivator.log(Level.INFO, "No configuration found. Server will not be running.", null);
			}
			else {
				FTPActivator.log(Level.INFO, "Reading configuration on server "+serverNameAbbr, null);
				
				String passivePorts=configDoc.getItemValueString("PassivePorts");
				if (passivePorts!=null && passivePorts.length()>0) {
					m_passivePorts=new PassivePorts(passivePorts, true);
				}
				String passiveAddress=configDoc.getItemValueString("PassiveAddress");
				if (passiveAddress!=null && passiveAddress.length()>0) {
					m_passiveAddress=passiveAddress;
				}
				String passiveExternalAddress=configDoc.getItemValueString("PassiveExternalAddress");
				if (passiveExternalAddress!=null && passiveExternalAddress.length()>0) {
					m_passiveExternalAddress=passiveExternalAddress;
				}
				int idleTime=configDoc.getItemValueInteger("IdleTime");
				if (idleTime!=0) {
					m_idleTime=Integer.valueOf(idleTime);
				}
				int ftpPort=configDoc.getItemValueInteger("Port");
				if (ftpPort!=0)
					m_ftpPort=Integer.valueOf(ftpPort);
				if ("1".equals(configDoc.getItemValueString("Enabled")))
					m_enabled=true;
				
				//parse blocked addresses
				m_blockedAddresses=getBlockedAddresses(configDoc);
				
				//parse blocked subnets
				m_blockedSubnets=getBlockedSubnets(configDoc);
				
				if (m_blockedAddresses.isEmpty() && m_blockedSubnets.isEmpty()) {
					FTPActivator.log(Level.INFO, "Configuration read successfully. No IP's or subnets blocked.", null);
				}
				else {
					FTPActivator.log(Level.INFO, "Configuration read successfully. Blocked IP's: "+m_blockedAddresses+", blocked subnets: "+m_blockedSubnets, null);
				}
			}
		}
		catch (NotesException e) {
			FTPActivator.log(Level.SEVERE, "Error reading FTP server configuration", e);
		}
		finally {
			if (session!=null) {
				try {
					session.recycle();
				}
				catch (NotesException e) {
					//
				}
			}
			NotesThread.stermThread();
		}
	}

	/**
	 * Reads the list of blocked IP's from the config document
	 * 
	 * @param configDoc config document
	 * @return list
	 * @throws NotesException
	 */
	private List<InetAddress> getBlockedAddresses(Document configDoc) throws NotesException {
		List<InetAddress> blockedAddresses=new ArrayList<InetAddress>();
		Vector blockedAddressesStrList=configDoc.getItemValue("BlockedAddresses");
		if (blockedAddressesStrList!=null && !blockedAddressesStrList.isEmpty()) {
			for (int i=0; i<blockedAddressesStrList.size(); i++) {
				String currAddrStr=blockedAddressesStrList.get(i).toString().trim();
				if ("".equals(currAddrStr))
					continue;
				
				try {
					InetAddress currAddr=InetAddress.getByName(currAddrStr);
					//try to resolve hostname
					currAddr.getHostAddress();
					blockedAddresses.add(currAddr);
				}
				catch (UnknownHostException e) {
					FTPActivator.log(Level.SEVERE, "Internet address to block has wrong format or cannot be resolved: "+currAddrStr, e);
				}
			}
		}
		return blockedAddresses;
	}
	
	/**
	 * Reads the list of blocked subnets from the config document
	 * 
	 * @param configDoc config document
	 * @return list
	 * @throws NotesException
	 */
	private List<Subnet> getBlockedSubnets(Document configDoc) throws NotesException {
		List<Subnet> blockedSubnets=new ArrayList<Subnet>();
		Vector blockedSubnetsStrList=configDoc.getItemValue("BlockedSubnets");
		if (blockedSubnetsStrList!=null && !blockedSubnetsStrList.isEmpty()) {
			for (int i=0; i<blockedSubnetsStrList.size(); i++) {
				String currSubnetStr=blockedSubnetsStrList.get(i).toString().trim();
				if ("".equals(currSubnetStr))
					continue;
				
				int iPos=currSubnetStr.indexOf("/");
				if (iPos==-1) {
					FTPActivator.log(Level.SEVERE, "Missing delimiter / in subnet to block: "+currSubnetStr, null);
				}
				else {
					String currHost=iPos==0 ? "" : currSubnetStr.substring(0, iPos);
					String currMaskStr=currSubnetStr.substring(iPos+1);
					InetAddress subnetInetAddr = null;
					try {
						subnetInetAddr=InetAddress.getByName(currHost);
						subnetInetAddr.getHostName();
					}
					catch (UnknownHostException e) {
						FTPActivator.log(Level.SEVERE, "Internet address of subnet to block has wrong format or cannot be resolved: "+currSubnetStr, e);
					}
					
					if (subnetInetAddr!=null) {
						try {
							int currMask=Integer.parseInt(currMaskStr);
							Subnet subnet=new Subnet(subnetInetAddr, currMask);
							blockedSubnets.add(subnet);
						}
						catch (NumberFormatException e) {
							FTPActivator.log(Level.SEVERE, "Bitmask of subnet to block is not a number: "+currSubnetStr, e);
						}
					}
				}
			}
		}
		return blockedSubnets;
	}
	
	/**
	 * Checks if an IP is blocked according to the blocked networks lists
	 * 
	 * @param inetAddr IP
	 * @return true if blocked
	 */
	public synchronized boolean isBlocked(InetAddress inetAddr) {
		for (InetAddress currAddr : m_blockedAddresses) {
			if (currAddr.equals(inetAddr))
				return true;
		}
		for (Subnet currSubnet : m_blockedSubnets) {
			if (currSubnet.inSubnet(inetAddr))
				return true;
		}
		return false;
	}
	
	/**
	 * Re-reads the list of blocked IP's and subnets from the FTP server configuration
	 */
	public synchronized void reloadBlockedNetworks() {
		List<InetAddress> blockedAddresses=new ArrayList<InetAddress>();
		List<Subnet> blockedSubnets=new ArrayList<Subnet>();
		
		NotesThread.sinitThread();
		Session session = null;
		try {
			session=NotesFactory.createSession();
			Database configDb=FTPDbLocator.getConfigDatabase(session);
			if (configDb==null) {
				FTPActivator.log(Level.WARNING, "No FTP server configuration database found!", null);
				return;
			}
			View lkView=configDb.getView("($lkConfigByServer)");
			if (lkView==null) {
				FTPActivator.log(Level.WARNING, "View ($lkConfigByServer) not found in FTP server configuration database", null);
				return;
			}

			String serverNameAsString=session.getServerName();
			Name serverName=session.createName(serverNameAsString);
			String serverNameAbbr=serverName.getAbbreviated();
			serverName.recycle();
			Document configDoc=lkView.getDocumentByKey(serverNameAbbr, true);
			if (configDoc!=null && configDoc.isValid()) {
				//parse blocked addresses
				blockedAddresses=getBlockedAddresses(configDoc);
				
				//parse blocked subnets
				blockedSubnets=getBlockedSubnets(configDoc);
				
				//write log entries if list has changed
				boolean blockedAddressesChanged=!m_blockedAddresses.equals(blockedAddresses);
				boolean blockedSubnetsChanged=!m_blockedSubnets.equals(blockedSubnets);
				
				if (blockedAddressesChanged) {
					m_blockedAddresses=blockedAddresses;
					FTPActivator.log(Level.INFO, "Read new list of blocked IPs: "+blockedAddresses, null);
				}
				if (blockedSubnetsChanged) {
					m_blockedSubnets=blockedSubnets;
					FTPActivator.log(Level.INFO, "Read new list of blocked subnets: "+blockedSubnets, null);
				}
			}
		}
		catch (NotesException e) {
			FTPActivator.log(Level.SEVERE, "Error reading blocked network list from FTP server configuration", e);
		}
		finally {
			if (session!=null) {
				try {
					session.recycle();
				}
				catch (NotesException e) {
					//
				}
			}
			NotesThread.stermThread();
		}
	}

}
